# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Train models with flexible insertion of algorithms."""

from composer.trainer.trainer import Trainer

__all__ = ['Trainer']
